"""
Methods and classes to distribute jobs computing Shapley values in a cluster.

You probably aren't interested in any of this unless you are developing new
methods for pyDVL that use parallelization.
"""

import logging
import operator
from functools import reduce
from time import time
from typing import cast

import numpy as np

from pydvl.utils.config import ParallelConfig
from pydvl.utils.parallel import init_parallel_backend
from pydvl.utils.parallel.actor import Coordinator, RayActorWrapper, Worker
from pydvl.utils.utility import Utility
from pydvl.value.result import ValuationResult
from pydvl.value.shapley.truncated import TruncationPolicy
from pydvl.value.stopping import MaxChecks, StoppingCriterion

__all__ = ["get_shapley_coordinator", "get_shapley_worker"]


logger = logging.getLogger(__name__)


def get_shapley_coordinator(
    *args, config: ParallelConfig = ParallelConfig(), **kwargs
) -> "ShapleyCoordinator":
    if config.backend == "ray":
        coordinator = cast(
            ShapleyCoordinator,
            RayActorWrapper(ShapleyCoordinator, config, *args, **kwargs),
        )
    elif config.backend == "sequential":
        coordinator = ShapleyCoordinator(*args, **kwargs)
    else:
        raise NotImplementedError(f"Unexpected parallel type {config.backend}")
    return coordinator


def get_shapley_worker(
    u: Utility, *args, config: ParallelConfig = ParallelConfig(), **kwargs
) -> "ShapleyWorker":
    parallel_backend = init_parallel_backend(config)
    u_id = parallel_backend.put(u)
    if config.backend == "ray":
        worker = cast(
            ShapleyWorker, RayActorWrapper(ShapleyWorker, config, u_id, *args, **kwargs)
        )
    elif config.backend == "sequential":
        worker = ShapleyWorker(u_id, *args, **kwargs)
    else:
        raise NotImplementedError(f"Unexpected parallel type {config.backend}")
    return worker


class ShapleyCoordinator(Coordinator):
    """The coordinator has two main tasks: aggregating the results of the
    workers and terminating processes once a certain stopping criterion is
    satisfied.
    """

    def __init__(self, done: StoppingCriterion):
        super().__init__()
        self.results_done = done
        self.results_done.modify_result = True

    def accumulate(self) -> ValuationResult:
        """Accumulates all results received from the workers.

        :return: Values and standard errors in a
            :class:`~pydvl.value.result.ValuationResult`. If no worker has
            reported yet, returns ``None``.
        """
        if len(self.worker_results) == 0:
            return ValuationResult.empty()  # type: ignore

        # FIXME: inefficient, possibly unstable
        totals: ValuationResult = reduce(operator.add, self.worker_results)
        # Avoid recomputing
        self.worker_results = [totals]
        return totals

    def check_convergence(self) -> bool:
        """Evaluates the convergence criterion on the accumulated results.

        If the convergence criterion is satisfied, calls to
        :meth:`~Coordinator.is_done` return ``True``.

        :return: ``True`` if converged and ``False`` otherwise.
        """
        if self.is_done():
            return True
        if len(self.worker_results) > 0:
            self._status = self.results_done(self.accumulate())
        return self.is_done()


class ShapleyWorker(Worker):
    """A worker.

    It should work.
    """

    algorithm: str = "truncated_montecarlo_shapley"

    def __init__(
        self,
        u: Utility,
        coordinator: ShapleyCoordinator,
        *,
        truncation: TruncationPolicy,
        worker_id: int,
        update_period: int = 30,
    ):
        """A worker calculates Shapley values using the permutation definition
         and reports the results to the coordinator.

         To implement early stopping, workers can be signaled by the
         :class:`~pydvl.value.shapley.actor.ShapleyCoordinator` before they are
         done with their work package

        :param u: Utility object with model, data, and scoring function
        :param coordinator: worker results will be pushed to this coordinator
        :param worker_id: id used for reporting through maybe_progress
        :param update_period: interval in seconds between different updates to
            and from the coordinator
        :param truncation: callable that decides whether to stop computing
            marginals for a given permutation.
        """
        super().__init__(
            coordinator=coordinator, update_period=update_period, worker_id=worker_id
        )
        self.u = u
        self.truncation = truncation

    def _compute_marginals(self) -> ValuationResult:
        # Avoid circular imports
        from .montecarlo import _permutation_montecarlo_shapley

        return _permutation_montecarlo_shapley(
            self.u,
            done=MaxChecks(1),
            truncation=self.truncation,
            algorithm_name=self.algorithm,
        )

    def run(self, *args, **kwargs):
        """Computes marginal utilities in a loop until signalled to stop.

        This calls :meth:`_compute_marginals` repeatedly calculating Shapley
        values on different permutations of the indices. After :attr:`update_period`
        seconds have passed, it reports the results to the
        :class:`~pydvl.value.shapley.actor.ShapleyCoordinator`. Before starting
        the next iteration, it checks the coordinator's
        :meth:`~pydvl.utils.parallel.actor.Coordinator.is_done` flag,
        terminating if it's ``True``.
        """
        while True:
            acc = ValuationResult.empty()
            start_time = time()
            while (time() - start_time) < self.update_period:
                if self.coordinator.is_done():
                    return
                results = self._compute_marginals()
                nans = np.isnan(results.values).sum()
                if nans > 0:
                    logger.warning(
                        f"{nans} NaN values in current permutation, ignoring. "
                        "Consider setting a default value for the Scorer"
                    )
                    continue
                acc += results
            self.coordinator.add_results(acc)



"""
This module contains Shapley computations for K-Nearest Neighbours.

.. todo::
   Implement approximate KNN computation for sublinear complexity)
"""

from typing import Dict, Union

import numpy as np
from sklearn.neighbors import KNeighborsClassifier, NearestNeighbors

from pydvl.utils import Utility, maybe_progress
from pydvl.utils.status import Status
from pydvl.value.result import ValuationResult

__all__ = ["knn_shapley"]


def knn_shapley(u: Utility, *, progress: bool = True) -> ValuationResult:
    """Computes exact Shapley values for a KNN classifier.

    This implements the method described in :footcite:t:`jia_efficient_2019a`.
    It exploits the local structure of K-Nearest Neighbours to reduce the number
    of calls to the utility function to a constant number per index, thus
    reducing computation time to $O(n)$.

    :param u: Utility with a KNN model to extract parameters from. The object
        will not be modified nor used other than to call `get_params()
        <https://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html#sklearn.base.BaseEstimator.get_params>`_
    :param progress: Whether to display a progress bar.
    :return: Object with the data values.
    :raises TypeError: If the model in the utility is not a `KNeighborsClassifier
        <https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html>`_

    .. versionadded:: 0.1.0

    """
    if not isinstance(u.model, KNeighborsClassifier):
        raise TypeError("KNN Shapley requires a K-Nearest Neighbours model")

    defaults: Dict[str, Union[int, str]] = {
        "algorithm": "ball_tree" if u.data.dim >= 20 else "kd_tree",
        "metric": "minkowski",
        "p": 2,
    }
    defaults.update(u.model.get_params())
    # HACK: NearestNeighbors doesn't support this. There will be more...
    del defaults["weights"]
    n_neighbors: int = int(defaults["n_neighbors"])
    defaults["n_neighbors"] = len(u.data)  # We want all training points sorted

    assert n_neighbors < len(u.data)
    # assert data.target_dim == 1

    nns = NearestNeighbors(**defaults).fit(u.data.x_train)
    # closest to farthest
    _, indices = nns.kneighbors(u.data.x_test)

    values = np.zeros_like(u.data.indices, dtype=np.float_)
    n = len(u.data)
    yt = u.data.y_train
    iterator = enumerate(zip(u.data.y_test, indices), start=1)
    for j, (y, ii) in maybe_progress(iterator, progress):
        value_at_x = int(yt[ii[-1]] == y) / n
        values[ii[-1]] += (value_at_x - values[ii[-1]]) / j
        for i in range(n - 2, n_neighbors, -1):  # farthest to closest
            value_at_x = (
                values[ii[i + 1]] + (int(yt[ii[i]] == y) - int(yt[ii[i + 1]] == y)) / i
            )
            values[ii[i]] += (value_at_x - values[ii[i]]) / j
        for i in range(n_neighbors, -1, -1):  # farthest to closest
            value_at_x = (
                values[ii[i + 1]]
                + (int(yt[ii[i]] == y) - int(yt[ii[i + 1]] == y)) / n_neighbors
            )
            values[ii[i]] += (value_at_x - values[ii[i]]) / j

    return ValuationResult(
        algorithm="knn_shapley",
        status=Status.Converged,
        values=values,
        data_names=u.data.data_names,
    )



import abc
import logging
from time import sleep

import numpy as np

from pydvl.utils import ParallelConfig, Utility, effective_n_jobs, running_moments
from pydvl.value import ValuationResult
from pydvl.value.stopping import StoppingCriterion

__all__ = [
    "TruncationPolicy",
    "NoTruncation",
    "FixedTruncation",
    "BootstrapTruncation",
    "RelativeTruncation",
    "truncated_montecarlo_shapley",
]


logger = logging.getLogger(__name__)


class TruncationPolicy(abc.ABC):
    """A policy for deciding whether to stop computing marginals in a
    permutation.

    Statistics are kept on the number of calls and truncations as :attr:`n_calls`
    and :attr:`n_truncations` respectively.

    .. todo::
       Because the policy objects are copied to the workers, the statistics
       are not accessible from the
       :class:`~pydvl.value.shapley.actor.ShapleyCoordinator`. We need to add
       methods for this.
    """

    def __init__(self):
        self.n_calls: int = 0
        self.n_truncations: int = 0

    @abc.abstractmethod
    def _check(self, idx: int, score: float) -> bool:
        """Implement the policy."""
        ...

    @abc.abstractmethod
    def reset(self):
        """Reset the policy to a state ready for a new permutation."""
        ...

    def __call__(self, idx: int, score: float) -> bool:
        """Check whether the computation should be interrupted.

        :param idx: Position in the permutation currently being computed.
        :param score: Last utility computed.
        :return: ``True`` if the computation should be interrupted.
        """
        ret = self._check(idx, score)
        self.n_calls += 1
        self.n_truncations += 1 if ret else 0
        return ret


class NoTruncation(TruncationPolicy):
    """A policy which never interrupts the computation."""

    def _check(self, idx: int, score: float) -> bool:
        return False

    def reset(self):
        pass


class FixedTruncation(TruncationPolicy):
    """Break a permutation after computing a fixed number of marginals.

    :param u: Utility object with model, data, and scoring function
    :param fraction: Fraction of marginals in a permutation to compute before
        stopping (e.g. 0.5 to compute half of the marginals).
    """

    def __init__(self, u: Utility, fraction: float):
        super().__init__()
        if fraction <= 0 or fraction > 1:
            raise ValueError("fraction must be in (0, 1]")
        self.max_marginals = len(u.data) * fraction
        self.count = 0

    def _check(self, idx: int, score: float) -> bool:
        self.count += 1
        return self.count >= self.max_marginals

    def reset(self):
        self.count = 0


class RelativeTruncation(TruncationPolicy):
    """Break a permutation if the marginal utility is too low.

    This is called "performance tolerance" in :footcite:t:`ghorbani_data_2019`.

    :param u: Utility object with model, data, and scoring function
    :param rtol: Relative tolerance. The permutation is broken if the
        last computed utility is less than ``total_utility * rtol``.
    """

    def __init__(self, u: Utility, rtol: float):
        super().__init__()
        self.rtol = rtol
        logger.info("Computing total utility for permutation truncation.")
        self.total_utility = u(u.data.indices)

    def _check(self, idx: int, score: float) -> bool:
        return np.allclose(score, self.total_utility, rtol=self.rtol)

    def reset(self):
        pass


class BootstrapTruncation(TruncationPolicy):
    """Break a permutation if the last computed utility is close to the total
    utility, measured as a multiple of the standard deviation of the utilities.

    :param u: Utility object with model, data, and scoring function
    :param n_samples: Number of bootstrap samples to use to compute the variance
        of the utilities.
    :param sigmas: Number of standard deviations to use as a threshold.
    """

    def __init__(self, u: Utility, n_samples: int, sigmas: float = 1):
        super().__init__()
        self.n_samples = n_samples
        logger.info("Computing total utility for permutation truncation.")
        self.total_utility = u(u.data.indices)
        self.count: int = 0
        self.variance: float = 0
        self.mean: float = 0
        self.sigmas: float = sigmas

    def _check(self, idx: int, score: float) -> bool:
        self.mean, self.variance = running_moments(
            self.mean, self.variance, self.count, score
        )
        self.count += 1
        logger.info(
            f"Bootstrap truncation: {self.count} samples, {self.variance:.2f} variance"
        )
        if self.count < self.n_samples:
            return False
        return abs(score - self.total_utility) < float(
            self.sigmas * np.sqrt(self.variance)
        )

    def reset(self):
        self.count = 0
        self.variance = self.mean = 0


def truncated_montecarlo_shapley(
    u: Utility,
    *,
    done: StoppingCriterion,
    truncation: TruncationPolicy,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    coordinator_update_period: int = 10,
    worker_update_period: int = 5,
) -> ValuationResult:
    """Monte Carlo approximation to the Shapley value of data points.

    This implements the permutation-based method described in
    :footcite:t:`ghorbani_data_2019`. It is a Monte Carlo estimate of the sum
    over all possible permutations of the index set, with a double stopping
    criterion.

    .. todo::
       Think of how to add Robin-Gelman or some other more principled stopping
       criterion.

    Instead of naively implementing the expectation, we sequentially add points
    to a dataset from a permutation and incrementally compute marginal utilities.
    We stop computing marginals for a given permutation based on a
    :class:`TruncationPolicy`. :footcite:t:`ghorbani_data_2019` mention two
    policies: one that stops after a certain fraction of marginals are computed,
    implemented in :class:`FixedTruncation`, and one that stops if the last
    computed utility ("score") is close to the total utility using the standard
    deviation of the utility as a measure of proximity, implemented in
    :class:`BootstrapTruncation`.

    We keep sampling permutations and updating all shapley values
    until the :class:`StoppingCriterion` returns ``True``.

    :param u: Utility object with model, data, and scoring function
    :param done: Check on the results which decides when to stop
        sampling permutations.
    :param truncation: callable that decides whether to stop computing
        marginals for a given permutation.
    :param n_jobs: number of jobs processing permutations. If None, it will be
        set to :func:`available_cpus`.
    :param config: Object configuring parallel computation, with cluster
        address, number of cpus, etc.
    :param coordinator_update_period: in seconds. How often to check the
        accumulated results from the workers for convergence.
    :param worker_update_period: interval in seconds between different
        updates to and from the coordinator
    :return: Object with the data values.

    """
    # Avoid circular imports
    from .actor import get_shapley_coordinator, get_shapley_worker

    if config.backend == "sequential":
        raise NotImplementedError(
            "Truncated MonteCarlo Shapley does not work with "
            "the Sequential parallel backend."
        )

    coordinator = get_shapley_coordinator(config=config, done=done)  # type: ignore

    workers = [
        get_shapley_worker(  # type: ignore
            u,
            coordinator=coordinator,
            truncation=truncation,
            worker_id=worker_id,
            update_period=worker_update_period,
            config=config,
        )
        for worker_id in range(effective_n_jobs(n_jobs, config=config))
    ]
    for worker in workers:
        worker.run(block=False)

    while not coordinator.check_convergence():
        sleep(coordinator_update_period)

    return coordinator.accumulate()

    # Something like this would be nicer, but it doesn't seem to be possible
    # to start the workers from the coordinator.
    # coordinator.add_workers(
    #     n_workers=n_jobs,
    #     u=u_id,
    #     update_period=worker_update_period,
    #     config=config,
    #     truncation=truncation,
    # )
    #
    # return coordinator.run(delay=coordinator_update_period)



"""
Samplers iterate over subsets of indices.

The classes in this module are used to iterate over indices and subsets of their
complement in the whole set, as required for the computation of marginal utility
for semi-values. The elements returned when iterating over any subclass of
:class:`PowersetSampler` are tuples of the form ``(idx, subset)``, where ``idx``
is the index of the element being added to the subset, and ``subset`` is the
subset of the complement of ``idx``.

.. note::
   This is the natural mode of iteration for the combinatorial definition of
   semi-values, in particular Shapley value. For the computation using
   permutations, adhering to this interface is not ideal, but we stick to it for
   consistency.

The samplers are used in the :mod:`pydvl.value.semivalues` module to compute any
semi-value, in particular Shapley and Beta values, and Banzhaf indices.

.. rubric:: Slicing of samplers

The samplers can be sliced for parallel computation. For those which are
embarrassingly parallel, this is done by slicing the set of "outer" indices and
returning new samplers over those slices. This includes all truly powerset-based
samplers, such as :class:`DeterministicCombinatorialSampler` and
:class:`UniformSampler`. In contrast, slicing a :class:`PermutationSampler`
creates a new sampler which iterates over the same indices.
"""

from __future__ import annotations

import abc
import math
from enum import Enum
from itertools import permutations
from typing import Generic, Iterable, Iterator, Sequence, Tuple, TypeVar, overload

import numpy as np
from numpy.typing import NDArray

from pydvl.utils.numeric import powerset, random_subset, random_subset_of_size

__all__ = [
    "AntitheticSampler",
    "DeterministicCombinatorialSampler",
    "DeterministicPermutationSampler",
    "PermutationSampler",
    "PowersetSampler",
    "RandomHierarchicalSampler",
    "UniformSampler",
]

T = TypeVar("T", bound=np.generic)
SampleType = Tuple[T, NDArray[T]]
Sequence.register(np.ndarray)


class PowersetSampler(abc.ABC, Iterable[SampleType], Generic[T]):
    """Samplers iterate over subsets of indices.

    This is done in nested loops, where the outer loop iterates over the set of
    indices, and the inner loop iterates over subsets of the complement of the
    current index. The outer iteration can be either sequential or at random.

    :Example:

    >>>for idx, s in DeterministicCombinatorialSampler([1,2]):
    >>>    print(s, end="")
    ()(2,)()(1,)

    .. rubric:: Methods required in subclasses

    Samplers must define a :meth:`weight` function to be used as a multiplier in
    Monte Carlo sums, so that the limit expectation coincides with the
    semi-value.

    .. rubric:: Slicing of samplers

    The samplers can be sliced for parallel computation. For those which are
    embarrassingly parallel, this is done by slicing the set of "outer" indices
    and returning new samplers over those slices.
    """

    class IndexIteration(Enum):
        Sequential = "sequential"
        Random = "random"

    def __init__(
        self,
        indices: NDArray[T],
        index_iteration: IndexIteration = IndexIteration.Sequential,
        outer_indices: NDArray[T] = None,
    ):
        """
        :param indices: The set of items (indices) to sample from.
        :param index_iteration: the order in which indices are iterated over
        :param outer_indices: The set of items (indices) over which to iterate
            when sampling. Subsets are taken from the complement of each index
            in succession. For embarrassingly parallel computations, this set
            is sliced and the samplers are used to iterate over the slices.
        """
        self._indices = indices
        self._index_iteration = index_iteration
        self._outer_indices = outer_indices if outer_indices is not None else indices
        self._n = len(indices)
        self._n_samples = 0

    @property
    def indices(self) -> NDArray[T]:
        return self._indices

    @indices.setter
    def indices(self, indices: NDArray[T]):
        raise AttributeError("Cannot set indices of sampler")

    @property
    def n_samples(self) -> int:
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n: int):
        raise AttributeError("Cannot reset a sampler's number of samples")

    def complement(self, exclude: Sequence[T]) -> NDArray[T]:
        return np.setxor1d(self._indices, exclude)

    def iterindices(self) -> Iterator[T]:
        """Iterates over indices in the order specified at construction.

        FIXME: this is probably not very useful, but I couldn't decide
          which method is better
        """
        if self._index_iteration is PowersetSampler.IndexIteration.Sequential:
            for idx in self._outer_indices:
                yield idx
        elif self._index_iteration is PowersetSampler.IndexIteration.Random:
            while True:
                yield np.random.choice(self._outer_indices, size=1).item()

    @overload
    def __getitem__(self, key: slice) -> "PowersetSampler[T]":
        ...

    @overload
    def __getitem__(self, key: list[int]) -> "PowersetSampler[T]":
        ...

    def __getitem__(self, key: slice | list[int]) -> "PowersetSampler[T]":
        if isinstance(key, slice) or isinstance(key, Iterable):
            return self.__class__(
                self._indices,
                index_iteration=self._index_iteration,
                outer_indices=self._indices[key],
            )
        raise TypeError("Indices must be an iterable or a slice")

    def __len__(self) -> int:
        """Returns the number of outer indices over which the sampler iterates."""
        return len(self._outer_indices)

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self._indices})"

    @abc.abstractmethod
    def __iter__(self) -> Iterator[SampleType]:
        ...

    @abc.abstractmethod
    def weight(self, subset: NDArray[T]) -> float:
        r"""Factor by which to multiply Monte Carlo samples, so that the
        mean converges to the desired expression.

        By the Law of Large Numbers, the sample mean of $\delta_i(S_j)$
        converges
        to the expectation under the distribution from which $S_j$ is sampled.

        $$ \frac{1}{m}  \sum_{j = 1}^m \delta_i (S_j) c (S_j) \longrightarrow
           \underset{S \sim \mathcal{D}_{- i}}{\mathbb{E}} [\delta_i (S) c (
           S)]$$

        We add a factor $c(S_j)$ in order to have this expectation coincide with
        the desired expression.
        """
        ...


class DeterministicCombinatorialSampler(PowersetSampler[T]):
    def __init__(self, indices: NDArray[T], *args, **kwargs):
        """Uniform deterministic sampling of subsets.

        For every index $i$, each subset of `indices - {i}` has equal
        probability $2^{n-1}$.

        :param indices: The set of items (indices) to sample from.
        """
        # Force sequential iteration
        kwargs.update({"index_iteration": PowersetSampler.IndexIteration.Sequential})
        super().__init__(indices, *args, **kwargs)

    def __iter__(self) -> Iterator[SampleType]:
        for idx in self.iterindices():
            for subset in powerset(self.complement([idx])):
                yield idx, np.array(subset)
                self._n_samples += 1

    def weight(self, subset: NDArray[T]) -> float:
        return float(2 ** (self._n - 1)) if self._n > 0 else 1.0


class UniformSampler(PowersetSampler[T]):
    def __iter__(self) -> Iterator[SampleType]:
        while True:
            for idx in self.iterindices():
                subset = random_subset(self.complement([idx]))
                yield idx, subset
                self._n_samples += 1
            if self._n_samples == 0:  # Empty index set
                break

    def weight(self, subset: NDArray[T]) -> float:
        """Correction coming from Monte Carlo integration so that the mean of
        the marginals converges to the value: the uniform distribution over the
        powerset of a set with n-1 elements has mass 2^{n-1} over each subset.
        The factor 1 / n corresponds to the one in the Shapley definition."""
        return float(2 ** (self._n - 1)) if self._n > 0 else 1.0


class AntitheticSampler(PowersetSampler[T]):
    def __iter__(self) -> Iterator[SampleType]:
        while True:
            for idx in self.iterindices():
                subset = random_subset(self.complement([idx]))
                yield idx, subset
                self._n_samples += 1
                yield idx, self.complement(np.concatenate((subset, np.array([idx]))))
                self._n_samples += 1
            if self._n_samples == 0:  # Empty index set
                break

    def weight(self, subset: NDArray[T]) -> float:
        return float(2 ** (self._n - 1)) if self._n > 0 else 1.0


class PermutationSampler(PowersetSampler[T]):
    """Sample permutations of indices and iterate through each returning sets,
    as required for the permutation definition of semi-values.

    .. warning::
       This sampler requires caching to be enabled or computation
       will be doubled wrt. a "direct" implementation of permutation MC
    """

    def __iter__(self) -> Iterator[SampleType]:
        while True:
            permutation = np.random.permutation(self._indices)
            for i, idx in enumerate(permutation):
                yield idx, permutation[:i]
                self._n_samples += 1
            if self._n_samples == 0:  # Empty index set
                break

    def __getitem__(self, key: slice | list[int]) -> "PowersetSampler[T]":
        """Permutation samplers cannot be split across indices, so we return
        a copy of the full sampler."""
        return super().__getitem__(slice(None))

    def weight(self, subset: NDArray[T]) -> float:
        return self._n * math.comb(self._n - 1, len(subset)) if self._n > 0 else 1.0


class DeterministicPermutationSampler(PermutationSampler[T]):
    """Samples all n! permutations of the indices deterministically, and
    iterates through them, returning sets as required for the permutation-based
    definition of semi-values.

    .. warning::
       This sampler requires caching to be enabled or computation
       will be doubled wrt. a "direct" implementation of permutation MC
    """

    def __iter__(self) -> Iterator[SampleType]:
        for permutation in permutations(self._indices):
            for i, idx in enumerate(permutation):
                yield idx, np.array(permutation[:i], dtype=self._indices.dtype)
                self._n_samples += 1
            if self._n_samples == 0:  # Empty index set
                break


class RandomHierarchicalSampler(PowersetSampler[T]):
    """For every index, sample a set size, then a set of that size.

    .. todo::
       This is unnecessary, but a step towards proper stratified sampling.
    """

    def __iter__(self) -> Iterator[SampleType]:
        while True:
            for idx in self.iterindices():
                k = np.random.choice(np.arange(len(self._indices)), size=1).item()
                subset = random_subset_of_size(self.complement([idx]), size=k)
                yield idx, subset
                self._n_samples += 1
            if self._n_samples == 0:  # Empty index set
                break

    def weight(self, subset: NDArray[T]) -> float:
        return float(2 ** (self._n - 1)) if self._n > 0 else 1.0



"""
Contains all parts of pyTorch based machine learning model.
"""
from typing import TYPE_CHECKING, Callable, Optional, Tuple, Union

import numpy as np

from ...utils import maybe_progress
from ..types import TwiceDifferentiable

try:
    import torch
    import torch.nn as nn
    from torch import autograd
    from torch.autograd import Variable

    _TORCH_INSTALLED = True
except ImportError:
    _TORCH_INSTALLED = False

if TYPE_CHECKING:
    from numpy.typing import NDArray

__all__ = [
    "TorchTwiceDifferentiable",
]


def flatten_gradient(grad):
    """
    Simple function to flatten a pyTorch gradient for use in subsequent calculation
    """
    return torch.cat([el.reshape(-1) for el in grad])


class TorchTwiceDifferentiable(TwiceDifferentiable):
    """
    Calculates second-derivative matrix vector products (Mvp) of a pytorch torch.nn.Module
    """

    def __init__(
        self,
        model: "nn.Module",
        loss: Callable[["torch.Tensor", "torch.Tensor"], "torch.Tensor"],
    ):
        """
        :param model: A torch.nn.Module representing a (differentiable) function f(x).
        :param loss: Loss function L(f(x), y) maps a prediction and a target to a single value.
        """
        if not _TORCH_INSTALLED:
            raise RuntimeWarning("This function requires PyTorch.")

        self.model = model
        self.loss = loss

    def num_params(self) -> int:
        """
        Get number of parameters of model f.
        :returns: Number of parameters as integer.
        """
        model_parameters = filter(lambda p: p.requires_grad, self.model.parameters())
        return sum([np.prod(p.size()) for p in model_parameters])

    def split_grad(
        self,
        x: Union["NDArray", "torch.Tensor"],
        y: Union["NDArray", "torch.Tensor"],
        progress: bool = False,
    ) -> "NDArray":
        """
        Calculates gradient of model parameters wrt each x[i] and y[i] and then
        returns a array of size [N, P] with N number of points (length of x and y) and P
        number of parameters of the model.
        :param x: A np.ndarray [NxD] representing the features x_i.
        :param y: A np.ndarray [NxK] representing the predicted target values y_i.
        :param progress: True, iff progress shall be printed.
        :returns: A np.ndarray [NxP] representing the gradients with respect to all parameters of the model.
        """
        x = torch.as_tensor(x).unsqueeze(1)
        y = torch.as_tensor(y)

        params = [
            param for param in self.model.parameters() if param.requires_grad == True
        ]

        grads = [
            flatten_gradient(
                autograd.grad(
                    self.loss(
                        torch.squeeze(self.model(x[i])),
                        torch.squeeze(y[i]),
                    ),
                    params,
                )
            )
            .detach()
            .numpy()
            for i in maybe_progress(
                range(len(x)),
                progress,
                desc="Split Gradient",
            )
        ]
        return np.stack(grads, axis=0)

    def grad(
        self,
        x: Union["NDArray", "torch.Tensor"],
        y: Union["NDArray", "torch.Tensor"],
    ) -> Tuple["NDArray", "torch.Tensor"]:
        """
        Calculates gradient of model parameters wrt x and y.
        :param x: A np.ndarray [NxD] representing the features x_i.
        :param y: A np.ndarray [NxK] representing the predicted target values y_i.
        :param progress: True, iff progress shall be printed.
        :returns: A tuple where: \
            - first element is a np.ndarray [P] with the gradients of the model. \
            - second element is the input to the model as a grad parameters. \
                This can be used for further differentiation. 
        """
        x = torch.as_tensor(x).requires_grad_(True)
        y = torch.as_tensor(y)

        params = [
            param for param in self.model.parameters() if param.requires_grad == True
        ]

        loss_value = self.loss(torch.squeeze(self.model(x)), torch.squeeze(y))
        grad_f = torch.autograd.grad(loss_value, params, create_graph=True)
        return flatten_gradient(grad_f), x

    def mvp(
        self,
        grad_xy: Union["NDArray", "torch.Tensor"],
        v: Union["NDArray", "torch.Tensor"],
        progress: bool = False,
        backprop_on: Optional["torch.Tensor"] = None,
    ) -> "NDArray":
        """
        Calculates second order derivative of the model along directions v.
        This second order derivative can be on the model parameters or on another input parameter, 
        selected via the backprop_on argument.

        :param grad_xy: an array [P] holding the gradients of the model parameters wrt input x and labels y, \
            where P is the number of parameters of the model. It is typically obtained through self.grad.
        :param v: A np.ndarray [DxP] or a one dimensional np.array [D] which multiplies the Hessian, \
            where D is the number of directions.
        :param progress: True, iff progress shall be printed.
        :param backprop_on: tensor used in the second backpropagation (the first one is along x and y as defined \
            via grad_xy). If None, the model parameters are used.
        :returns: A np.ndarray representing the implicit matrix vector product of the model along the given directions.\
            Output shape is [DxP] if backprop_on is None, otherwise [DxM], with M the number of elements of backprop_on.
        """
        v = torch.as_tensor(v)
        if v.ndim == 1:
            v = v.unsqueeze(0)

        z = (grad_xy * Variable(v)).sum(dim=1)
        params = [
            param for param in self.model.parameters() if param.requires_grad == True
        ]
        all_flattened_grads = [
            flatten_gradient(
                autograd.grad(
                    z[i],
                    params if backprop_on is None else backprop_on,
                    retain_graph=True,
                )
            )
            for i in maybe_progress(
                range(len(z)),
                progress,
                desc="MVP",
            )
        ]
        hvp = torch.stack([grad.contiguous().view(-1) for grad in all_flattened_grads])
        return hvp.detach().numpy()  # type: ignore



"""
This module contains classes to manage and learn utility functions for the
computation of values. Please see the documentation on :ref:`data valuation` for
more information.

:class:`Utility` holds information about model, data and scoring function (the
latter being what one usually understands under *utility* in the general
definition of Shapley value). It is automatically cached across machines.

:class:`DataUtilityLearning` adds support for learning the scoring function
to avoid repeated re-training of the model to compute the score.

This module also contains Utility classes for toy games that are used
for testing and for demonstration purposes.

"""
import logging
import warnings
from dataclasses import asdict
from typing import Dict, FrozenSet, Iterable, Optional, Tuple, Union, cast

import numpy as np
from numpy.typing import NDArray
from sklearn.base import clone
from sklearn.metrics import check_scoring

from pydvl.utils import Dataset
from pydvl.utils.caching import CacheStats, memcached, serialize
from pydvl.utils.config import MemcachedConfig
from pydvl.utils.score import Scorer
from pydvl.utils.types import SupervisedModel

__all__ = ["Utility", "DataUtilityLearning", "MinerGameUtility", "GlovesGameUtility"]

logger = logging.getLogger(__name__)


class Utility:
    """Convenience wrapper with configurable memoization of the scoring
    function.

    An instance of ``Utility`` holds the triple of model, dataset and scoring
    function which determines the value of data points. This is mosly used for
    the computation of :ref:`Shapley values<Shapley>` and
    :ref:`Least Core values<Least Core>`.

    The Utility expect the model to fulfill
    the :class:`pydvl.utils.types.SupervisedModel` interface i.e. to have
    ``fit()``, ``predict()``, and ``score()`` methods.

    When calling the utility, the model will be
    `cloned <https://scikit-learn.org/stable/modules/generated/sklearn.base
    .clone.html>`_
    if it is a Sci-Kit Learn model, otherwise a copy is created using
    ``deepcopy()``
    from the builtin `copy <https://docs.python.org/3/library/copy.html>`_
    module.

    Since evaluating the scoring function requires retraining the model
    and that can be time-consuming, this class wraps it and caches
    the results of each execution. Caching is available both locally
    and across nodes, but must always be enabled for your
    project first, see :ref:`how to set up the cache<caching setup>`.

    :param model: Any supervised model. Typical choices can be found at
            https://scikit-learn.org/stable/supervised_learning.html
    :param data: :class:`Dataset` or :class:`GroupedDataset`.
    :param scorer: A scoring object. If None, the ``score()`` method of the model
        will be used. See :mod:`~pydvl.utils.scorer` for ways to create
        and compose scorers, in particular how to set default values and ranges.
        For convenience, a string can be passed, which will be used to construct
        a :class:`~pydvl.utils.scorer.Scorer`.
    :param default_score: As a convenience when no ``scorer`` object is passed
        (where a default value can be provided), this argument also allows to set
        the default score for models that have not been fit, e.g. when too little
        data is passed, or errors arise.
    :param score_range: As with ``default_score``, this is a convenience argument
        for when no ``scorer`` argument is provided, to set the numerical range
        of the score function. Some Monte Carlo methods can use this to estimate
        the number of samples required for a certain quality of approximation.
    :param catch_errors: set to ``True`` to catch the errors when fit() fails.
        This could happen in several steps of the pipeline, e.g. when too little
        training data is passed, which happens often during Shapley value
        calculations. When this happens, the :attr:`default_score` is returned
        as a score and computation continues.
    :param show_warnings: Set to ``False`` to suppress warnings thrown by
        ``fit()``.
    :param enable_cache: If ``True``, use memcached for memoization.
    :param cache_options: Optional configuration object for memcached.
    :param clone_before_fit: If True, the model will be cloned before calling
        ``fit()``.

    :Example:

    >>> from pydvl.utils import Utility, DataUtilityLearning, Dataset
    >>> from sklearn.linear_model import LinearRegression, LogisticRegression
    >>> from sklearn.datasets import load_iris
    >>> dataset = Dataset.from_sklearn(load_iris(), random_state=16)
    >>> u = Utility(LogisticRegression(random_state=16), dataset)
    >>> u(dataset.indices)
    0.9

    """

    model: SupervisedModel
    data: Dataset
    scorer: Scorer

    def __init__(
        self,
        model: SupervisedModel,
        data: Dataset,
        scorer: Optional[Union[str, Scorer]] = None,
        *,
        default_score: float = 0.0,
        score_range: Tuple[float, float] = (-np.inf, np.inf),
        catch_errors: bool = True,
        show_warnings: bool = False,
        enable_cache: bool = False,
        cache_options: Optional[MemcachedConfig] = None,
        clone_before_fit: bool = True,
    ):
        self.model = self._clone_model(model)
        self.data = data
        if isinstance(scorer, str):
            scorer = Scorer(scorer, default=default_score, range=score_range)
        self.scorer = check_scoring(self.model, scorer)
        self.default_score = scorer.default if scorer is not None else default_score
        # TODO: auto-fill from known scorers ?
        self.score_range = scorer.range if scorer is not None else np.array(score_range)
        self.catch_errors = catch_errors
        self.show_warnings = show_warnings
        self.enable_cache = enable_cache
        self.cache_options: MemcachedConfig = cache_options or MemcachedConfig()
        self.clone_before_fit = clone_before_fit
        self._signature = serialize((hash(self.model), hash(data), hash(scorer)))
        self._initialize_utility_wrapper()

        # FIXME: can't modify docstring of methods. Instead, I could use a
        #  factory which creates the class on the fly with the right doc.
        # self.__call__.__doc__ = self._utility_wrapper.__doc__

    def _initialize_utility_wrapper(self):
        if self.enable_cache:
            # asdict() is recursive, but we want client_config to remain a dataclass
            options = asdict(self.cache_options)
            options["client_config"] = self.cache_options.client_config
            self._utility_wrapper = memcached(**options)(  # type: ignore
                self._utility, signature=self._signature
            )
        else:
            self._utility_wrapper = self._utility

    def __call__(self, indices: Iterable[int]) -> float:
        utility: float = self._utility_wrapper(frozenset(indices))
        return utility

    def _utility(self, indices: FrozenSet) -> float:
        """Clones the model, fits it on a subset of the training data
        and scores it on the test data.

        If the object is constructed with ``enable_cache = True``, results are
        memoized to avoid duplicate computation. This is useful in particular
        when computing utilities of permutations of indices or when randomly
        sampling from the powerset of indices.

        :param indices: a subset of valid indices for
            :attr:`~pydvl.utils.dataset.Dataset.x_train`. The type must be
            hashable for the caching to work, e.g. wrap the argument with
            `frozenset <https://docs.python.org/3/library/stdtypes.html#frozenset>`_
            (rather than `tuple` since order should not matter)
        :return: 0 if no indices are passed, :attr:`default_score`` if we fail
            to fit the model or the scorer returns `NaN`. Otherwise, the score
            of the model on the test data.
        """
        if not indices:
            return 0.0

        x_train, y_train = self.data.get_training_data(list(indices))
        x_test, y_test = self.data.get_test_data(list(indices))

        with warnings.catch_warnings():
            if not self.show_warnings:
                warnings.simplefilter("ignore")
            try:
                if self.clone_before_fit:
                    model = self._clone_model(self.model)
                else:
                    model = self.model
                model.fit(x_train, y_train)
                score = float(self.scorer(model, x_test, y_test))
                # Some scorers raise exceptions if they return NaNs, some might not
                if np.isnan(score):
                    warnings.warn("Scorer returned NaN", RuntimeWarning)
                    return self.default_score
                return score
            except Exception as e:
                if self.catch_errors:
                    warnings.warn(str(e), RuntimeWarning)
                    return self.default_score
                raise

    @staticmethod
    def _clone_model(model: SupervisedModel) -> SupervisedModel:
        """Clones the passed model to avoid the possibility
        of reusing a fitted estimator

        :param model: Any supervised model. Typical choices can be found at
            https://scikit-learn.org/stable/supervised_learning.html
        """
        try:
            model = clone(model)
        except TypeError:
            # This happens if the passed model is not an sklearn model
            # In this case, we just make a deepcopy of the model.
            model = clone(model, safe=False)
        model = cast(SupervisedModel, model)
        return model

    @property
    def signature(self):
        """Signature used for caching model results."""
        return self._signature

    @property
    def cache_stats(self) -> Optional[CacheStats]:
        """Cache statistics are gathered when cache is enabled.
        See :class:`~pydvl.utils.caching.CacheInfo` for all fields returned.
        """
        if self.enable_cache:
            return self._utility_wrapper.stats  # type: ignore
        return None

    def __getstate__(self):
        state = self.__dict__.copy()
        # Don't pickle _utility_wrapper
        state.pop("_utility_wrapper", None)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # Add _utility_wrapper back since it doesn't exist in the pickle
        self._initialize_utility_wrapper()


class DataUtilityLearning:
    """Implementation of Data Utility Learning algorithm
    :footcite:t:`wang_improving_2022`.

    This object wraps a :class:`~pydvl.utils.utility.Utility` and delegates
    calls to it, up until a given budget (number of iterations). Every tuple
    of input and output (a so-called *utility sample*) is stored. Once the
    budget is exhausted, `DataUtilityLearning` fits the given model to the
    utility samples. Subsequent calls will use the learned model to predict the
    utility instead of delegating.

    :param u: The :class:`~pydvl.utils.utility.Utility` to learn.
    :param training_budget: Number of utility samples to collect before fitting
        the given model
    :param model: A supervised regression model

    :Example:

    >>> from pydvl.utils import Utility, DataUtilityLearning, Dataset
    >>> from sklearn.linear_model import LinearRegression, LogisticRegression
    >>> from sklearn.datasets import load_iris
    >>> dataset = Dataset.from_sklearn(load_iris())
    >>> u = Utility(LogisticRegression(), dataset)
    >>> wrapped_u = DataUtilityLearning(u, 3, LinearRegression())
    ... # First 3 calls will be computed normally
    >>> for i in range(3):
    ...     _ = wrapped_u((i,))
    >>> wrapped_u((1, 2, 3)) # Subsequent calls will be computed using the fit model for DUL
    0.0

    """

    def __init__(
        self, u: Utility, training_budget: int, model: SupervisedModel
    ) -> None:
        self.utility = u
        self.training_budget = training_budget
        self.model = model
        self._current_iteration = 0
        self._is_model_fit = False
        self._utility_samples: Dict[FrozenSet, Tuple[NDArray[np.bool_], float]] = {}

    def _convert_indices_to_boolean_vector(self, x: Iterable[int]) -> NDArray[np.bool_]:
        boolean_vector = np.zeros((1, len(self.utility.data)), dtype=bool)
        if x is not None:
            boolean_vector[:, tuple(x)] = True
        return boolean_vector

    def __call__(self, indices: Iterable[int]) -> float:
        indices_boolean_vector = self._convert_indices_to_boolean_vector(indices)
        frozen_indices = frozenset(indices)
        if self._current_iteration < self.training_budget:
            utility = self.utility(frozen_indices)
            self._utility_samples[frozen_indices] = (indices_boolean_vector, utility)
            self._current_iteration += 1
        else:
            if not self._is_model_fit:
                X, y = zip(*self._utility_samples.values())
                X = np.vstack(X)
                y = np.asarray(y)
                self.model.fit(X, y)
                self._is_model_fit = True
            if frozen_indices in self._utility_samples:
                utility = self._utility_samples[frozen_indices][1]
            else:
                utility = self.model.predict(indices_boolean_vector).item()
        return utility

    @property
    def data(self) -> Dataset:
        """Returns the wrapped utility's :class:`~pydvl.utils.dataset.Dataset`."""
        return self.utility.data


class MinerGameUtility(Utility):
    r"""Toy game utility that is used for testing and demonstration purposes.

    Consider a group of n miners, who have discovered large bars of gold.

    If two miners can carry one piece of gold,
    then the payoff of a coalition $S$ is:

    $${
    v(S) = \left\{\begin{array}{lll}
    \mid S \mid / 2 & \text{, if} & \mid S \mid \text{ is even} \\
    ( \mid S \mid - 1)/2 & \text{, if} & \mid S \mid \text{ is odd}
    \end{array}\right.
    }$$

    If there are more than two miners and there is an even number of miners,
    then the core consists of the single payoff where each miner gets 1/2.

    If there is an odd number of miners, then the core is empty.

    Taken from: https://en.wikipedia.org/wiki/Core_(game_theory)

    :param n_miners: Number of miners that participate in the game.
    """

    def __init__(self, n_miners: int, **kwargs):
        if n_miners <= 2:
            raise ValueError(f"n_miners, {n_miners} should be > 2")
        self.n_miners = n_miners

        x = np.arange(n_miners)[..., np.newaxis]
        # The y values don't matter here
        y = np.zeros_like(x)

        self.data = Dataset(x_train=x, y_train=y, x_test=x, y_test=y)

    def __call__(self, indices: Iterable[int]) -> float:
        n = len(tuple(indices))
        if n % 2 == 0:
            return n / 2
        else:
            return (n - 1) / 2

    def _initialize_utility_wrapper(self):
        pass

    def exact_least_core_values(self) -> Tuple[NDArray[np.float_], float]:
        if self.n_miners % 2 == 0:
            values = np.array([0.5] * self.n_miners)
            subsidy = 0.0
        else:
            values = np.array(
                [(self.n_miners - 1) / (2 * self.n_miners)] * self.n_miners
            )
            subsidy = (self.n_miners - 1) / (2 * self.n_miners)
        return values, subsidy

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(n={self.n_miners})"


class GlovesGameUtility(Utility):
    r"""Toy game utility that is used for testing and demonstration purposes.

    In this game, some players have a left glove and others a right glove.
    Single gloves have a worth of zero while pairs have a worth of 1.

    The payoff of a coalition $S$ is:

    $${
    v(S) = \min( \mid S \cap L \mid, \mid S \cap R \mid )
    }$$

    Where $L$, respectively $R$, is the set of players with left gloves,
    respectively right gloves.

    :param left: Number of players with a left glove.
    :param right: Number of player with a right glove.

    """

    def __init__(self, left: int, right: int, **kwargs):
        self.left = left
        self.right = right

        x = np.empty(left + right)[..., np.newaxis]
        # The y values don't matter here
        y = np.zeros_like(x)

        self.data = Dataset(x_train=x, y_train=y, x_test=x, y_test=y)

    def __call__(self, indices: Iterable[int]) -> float:
        left_sum = float(np.sum(np.asarray(indices) < self.left))
        right_sum = float(np.sum(np.asarray(indices) >= self.left))
        return min(left_sum, right_sum)

    def _initialize_utility_wrapper(self):
        pass

    def exact_least_core_values(self) -> Tuple[NDArray[np.float_], float]:
        if self.left == self.right:
            subsidy = -0.5
            values = np.array([0.5] * (self.left + self.right))
        elif self.left < self.right:
            subsidy = 0.0
            values = np.array([1.0] * self.left + [0.0] * self.right)
        else:
            subsidy = 0.0
            values = np.array([0.0] * self.left + [1.0] * self.right)
        return values, subsidy

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(L={self.left}, R={self.right})"



import abc
import inspect
import logging
from time import sleep
from typing import Generic, List, Optional, Type, TypeVar, cast

from ..config import ParallelConfig
from ..status import Status
from .backend import RayParallelBackend, init_parallel_backend

__all__ = ["RayActorWrapper", "Coordinator", "Worker"]


logger = logging.getLogger(__name__)


class RayActorWrapper:
    """Wrapper to call methods of remote Ray actors as if they were local.

    Taken almost verbatim from:
    https://github.com/JaneliaSciComp/ray-janelia/blob/main/remote_as_local_wrapper.py

    :Example:

    >>> from pydvl.utils.parallel.backend import RayParallelBackend, init_parallel_backend
    >>> from pydvl.utils.config import ParallelConfig
    >>> from pydvl.utils.parallel.actor import RayActorWrapper
    >>> class Actor:
    ...     def __init__(self, x):
    ...         self.x = x
    ...
    ...     def get(self):
    ...         return self.x
    ...
    >>> config = ParallelConfig(backend="ray")
    >>> wrapped_actor = RayActorWrapper(Actor, config, 5)
    >>> wrapped_actor.get()
    5
    """

    def __init__(self, actor_class: Type, config: ParallelConfig, *args, **kwargs):
        parallel_backend = cast(RayParallelBackend, init_parallel_backend(config))
        remote_cls = parallel_backend.wrap(actor_class)
        self.actor_handle = remote_cls(*args, **kwargs)

        def remote_caller(method_name: str):
            # Wrapper for remote class' methods to mimic local calls
            def wrapper(
                *args, block: bool = True, timeout: Optional[float] = None, **kwargs
            ):
                obj_ref = getattr(self.actor_handle, method_name).remote(
                    *args, **kwargs
                )
                if block:
                    return parallel_backend.get(
                        obj_ref, timeout=timeout
                    )  # Block until called method returns.
                else:
                    return obj_ref  # Don't block and return a future.

            return wrapper

        for member in inspect.getmembers(actor_class):
            name = member[0]
            if not name.startswith("__"):
                # Wrap public methods for remote-as-local calls.
                setattr(self, name, remote_caller(name))


Result = TypeVar("Result")  # Avoids circular import with ValuationResult


class Coordinator(Generic[Result], abc.ABC):
    """The coordinator has two main tasks: aggregating the results of the
    workers and terminating the process once a certain accuracy or total
    number of iterations is reached.
    """

    _status: Status

    def __init__(self):
        self.worker_results: List[Result] = []
        self._status = Status.Pending

    def add_results(self, results: Result):
        """Used by workers to report their results. Stores the results directly
        into :attr:`worker_results`

        :param results: results of worker's calculations
        """
        self.worker_results.append(results)

    # this should be a @property, but with it ray.get messes up
    def is_done(self) -> bool:
        """Used by workers to check whether to terminate their process.

        :return: ``True`` if workers must terminate, ``False`` otherwise.
        """
        return bool(self._status)

    @abc.abstractmethod
    def accumulate(self) -> Result:
        """Aggregates the results of the different workers."""
        raise NotImplementedError()

    @abc.abstractmethod
    def check_convergence(self) -> bool:
        """Evaluates the convergence criteria on the aggregated results."""
        raise NotImplementedError()


class Worker(abc.ABC):
    """A worker, it should work."""

    def __init__(
        self,
        coordinator: Coordinator,
        worker_id: int,
        *,
        update_period: int = 30,
    ):
        """A worker

        :param coordinator: worker results will be pushed to this coordinator
        :param worker_id: id used for reporting through maybe_progress
        :param update_period: interval in seconds between different updates
            to and from the coordinator
        """
        super().__init__()
        self.worker_id = worker_id
        self.coordinator = coordinator
        self.update_period = update_period

    def run(self, *args, **kwargs):
        """Runs the worker."""
        raise NotImplementedError()



import logging
from dataclasses import dataclass, field
from typing import Iterable, Literal, Optional, Tuple, Union

from pymemcache.serde import PickleSerde

PICKLE_VERSION = 5  # python >= 3.8

__all__ = ["ParallelConfig", "MemcachedClientConfig", "MemcachedConfig"]


@dataclass
class ParallelConfig:
    """Configuration for parallel computation backend.

    :param backend: Type of backend to use. For now only 'ray' is supported.
    :param address: Address of existing remote or local cluster to use.
    :param n_local_workers: Number of workers (CPUs) to use when using a local ray cluster.
    :param logging_level: Logging level for the parallel backend's worker.
    """

    backend: Literal["sequential", "ray"] = "ray"
    address: Optional[Union[str, Tuple[str, int]]] = None
    n_local_workers: Optional[int] = None
    logging_level: int = logging.WARNING


@dataclass
class MemcachedClientConfig:
    """Configuration of the memcached client.

    :param server: A tuple of (IP|domain name, port).
    :param connect_timeout: How many seconds to wait before raising
        `ConnectionRefusedError` on failure to connect.
    :param timeout: seconds to wait for send or recv calls on the socket
        connected to memcached.
    :param no_delay: set the `TCP_NODELAY` flag, which may help with performance
        in some cases.
    :param serde: a serializer / deserializer ("serde"). The default
        `PickleSerde` should work in most cases. See `pymemcached's
        documentation
        <https://pymemcache.readthedocs.io/en/latest/apidoc/pymemcache.client.base.html#pymemcache.client.base.Client>`_
        for details.
    """

    server: Tuple[str, int] = ("localhost", 11211)
    connect_timeout: float = 1.0
    timeout: float = 1.0
    no_delay: bool = True
    serde: PickleSerde = PickleSerde(pickle_version=PICKLE_VERSION)


@dataclass
class MemcachedConfig:
    """Configuration for :func:`~pydvl.utils.caching.memcached`, providing
    memoization of function calls.

    Instances of this class are typically used as arguments for the construction
    of a :class:`~pydvl.utils.utility.Utility`.

    :param client_config: Configuration for the connection to the memcached
        server.
    :param time_threshold: computations taking less time than this many seconds
        are not cached.
    :param allow_repeated_evaluations: If `True`, repeated calls to a function
        with the same arguments will be allowed and outputs averaged until the
        running standard deviation of the mean stabilises below
        `rtol_stderr * mean`.
    :param rtol_stderr: relative tolerance for repeated evaluations. More
        precisely, :func:`~pydvl.utils.caching.memcached` will stop evaluating
        the function once the standard deviation of the mean is smaller than
        `rtol_stderr * mean`.
    :param min_repetitions: minimum number of times that a function evaluation
        on the same arguments is repeated before returning cached values. Useful
        for stochastic functions only. If the model training is very noisy, set
        this number to higher values to reduce variance.
    :param ignore_args: Do not take these keyword arguments into account when
        hashing the wrapped function for usage as key in memcached.
    """

    client_config: MemcachedClientConfig = field(default_factory=MemcachedClientConfig)
    time_threshold: float = 0.3
    allow_repeated_evaluations: bool = False
    rtol_stderr: float = 0.1
    min_repetitions: int = 3
    ignore_args: Optional[Iterable[str]] = None



"""
Contains parallelized influence calculation functions for general models.
"""
from enum import Enum
from typing import TYPE_CHECKING, Callable, Dict, Optional

import numpy as np
from scipy.sparse.linalg import LinearOperator

from ..utils import maybe_progress
from .conjugate_gradient import (
    batched_preconditioned_conjugate_gradient,
    conjugate_gradient,
)
from .frameworks import TorchTwiceDifferentiable
from .types import MatrixVectorProductInversionAlgorithm, TwiceDifferentiable

try:
    import torch
    import torch.nn as nn

    _TORCH_INSTALLED = True
except ImportError:
    _TORCH_INSTALLED = False

if TYPE_CHECKING:
    from numpy.typing import NDArray


__all__ = ["compute_influences", "InfluenceType", "InversionMethod"]


class InfluenceType(str, Enum):
    """
    Different influence types.
    """

    Up = "up"
    Perturbation = "perturbation"


class InversionMethod(str, Enum):
    """
    Different inversion methods types.
    """

    Direct = "direct"
    Cg = "cg"
    BatchedCg = "batched_cg"


def calculate_influence_factors(
    model: TwiceDifferentiable,
    x: "NDArray",
    y: "NDArray",
    x_test: "NDArray",
    y_test: "NDArray",
    inversion_func: MatrixVectorProductInversionAlgorithm,
    lam: float = 0,
    progress: bool = False,
) -> "NDArray":
    """
    Calculates the influence factors. For more info, see https://arxiv.org/pdf/1703.04730.pdf, paragraph 3.

    :param model: A model which has to implement the TwiceDifferentiable interface.
    :param x_train: A np.ndarray of shape [MxK] containing the features of the input data points.
    :param y_train: A np.ndarray of shape [MxL] containing the targets of the input data points.
    :param x_test: A np.ndarray of shape [NxK] containing the features of the test set of data points.
    :param y_test: A np.ndarray of shape [NxL] containing the targets of the test set of data points.
    :param inversion_func: function to use to invert the product of hvp (hessian vector product) and the gradient
        of the loss (s_test in the paper).
    :param lam: regularization of the hessian
    :param progress: If True, display progress bars.
    :returns: A np.ndarray of size (N, D) containing the influence factors for each dimension (D) and test sample (N).
    """
    if not _TORCH_INSTALLED:
        raise RuntimeWarning("This function requires PyTorch.")
    grad_xy, _ = model.grad(x, y)
    hvp = lambda v: model.mvp(grad_xy, v) + lam * v
    test_grads = model.split_grad(x_test, y_test, progress)
    return inversion_func(hvp, test_grads)


def _calculate_influences_up(
    model: TwiceDifferentiable,
    x: "NDArray",
    y: "NDArray",
    influence_factors: "NDArray",
    progress: bool = False,
) -> "NDArray":
    """
    Calculates the influence from the influence factors and the scores of the training points.
    Uses the upweighting method, as described in section 2.1 of https://arxiv.org/pdf/1703.04730.pdf

    :param model: A model which has to implement the TwiceDifferentiable interface.
    :param x_train: A np.ndarray of shape [MxK] containing the features of the input data points.
    :param y_train: A np.ndarray of shape [MxL] containing the targets of the input data points.
    :param influence_factors: np.ndarray containing influence factors
    :param progress: If True, display progress bars.
    :returns: A np.ndarray of size [NxM], where N is number of test points and M number of train points.
    """
    train_grads = model.split_grad(x, y, progress)
    return np.einsum("ta,va->tv", influence_factors, train_grads)  # type: ignore


def _calculate_influences_pert(
    model: TwiceDifferentiable,
    x: "NDArray",
    y: "NDArray",
    influence_factors: "NDArray",
    progress: bool = False,
) -> "NDArray":
    """
    Calculates the influence from the influence factors and the scores of the training points.
    Uses the perturbation method, as described in section 2.2 of https://arxiv.org/pdf/1703.04730.pdf

    :param model: A model which has to implement the TwiceDifferentiable interface.
    :param x_train: A np.ndarray of shape [MxK] containing the features of the input data points.
    :param y_train: A np.ndarray of shape [MxL] containing the targets of the input data points.
    :param influence_factors: np.ndarray containing influence factors
    :param progress: If True, display progress bars.
    :returns: A np.ndarray of size [NxMxP], where N is number of test points, M number of train points,
        and P the number of features.
    """
    all_pert_influences = []
    for i in maybe_progress(
        len(x),
        progress,
        desc="Influence Perturbation",
    ):
        grad_xy, tensor_x = model.grad(x[i : i + 1], y[i])
        perturbation_influences = model.mvp(
            grad_xy,
            influence_factors,
            backprop_on=tensor_x,
        )
        all_pert_influences.append(perturbation_influences.reshape((-1, *x[i].shape)))

    return np.stack(all_pert_influences, axis=1)


influence_type_function_dict = {
    "up": _calculate_influences_up,
    "perturbation": _calculate_influences_pert,
}


def compute_influences(
    model: "nn.Module",
    loss: Callable[["torch.Tensor", "torch.Tensor"], "torch.Tensor"],
    x: "NDArray",
    y: "NDArray",
    x_test: "NDArray",
    y_test: "NDArray",
    progress: bool = False,
    inversion_method: InversionMethod = InversionMethod.Direct,
    influence_type: InfluenceType = InfluenceType.Up,
    inversion_method_kwargs: Optional[Dict] = None,
    hessian_regularization: float = 0,
) -> "NDArray":
    """
    Calculates the influence of the training points j on the test points i. First it calculates
    the influence factors for all test points with respect to the training points, and then uses them to
    get the influences over the complete training set. Points with low influence values are (on average)
    less important for model training than points with high influences.

    :param model: A supervised model from a supported framework. Currently, only pytorch nn.Module is supported.
    :param loss: loss of the model, a callable that, given prediction of the model and real labels, returns a
        tensor with the loss value.
    :param x: model input for training
    :param y: input labels
    :param x_test: model input for testing
    :param y_test: test labels
    :param progress: whether to display progress bars.
    :param inversion_method: Set the inversion method to a specific one, can be 'direct' for direct inversion
        (and explicit construction of the Hessian) or 'cg' for conjugate gradient.
    :param influence_type: Which algorithm to use to calculate influences.
        Currently supported options: 'up' or 'perturbation'. For details refer to https://arxiv.org/pdf/1703.04730.pdf
    :param inversion_method_kwargs: kwargs for the inversion method selected.
        If using the direct method no kwargs are needed. If inversion_method='cg', the following kwargs can be passed:
        - rtol: relative tolerance to be achieved before terminating computation
        - max_iterations: maximum conjugate gradient iterations
        - max_step_size: step size of conjugate gradient
        - verify_assumptions: True to run tests on convexity of the model.
    :param hessian_regularization: lambda to use in Hessian regularization, i.e. H_reg = H + lambda * 1, with 1 the identity matrix \
        and H the (simple and regularized) Hessian. Typically used with more complex models to make sure the Hessian \
        is positive definite.
    :returns: A np.ndarray specifying the influences. Shape is [NxM] if influence_type is'up', where N is number of test points and
        M number of train points. If instead influence_type is 'perturbation', output shape is [NxMxP], with P the number of input
        features.
    """
    if not _TORCH_INSTALLED:
        raise RuntimeWarning("This function requires PyTorch.")

    if inversion_method_kwargs is None:
        inversion_method_kwargs = dict()
    differentiable_model = TorchTwiceDifferentiable(model, loss)
    n_params = differentiable_model.num_params()
    dict_fact_algos: Dict[Optional[str], MatrixVectorProductInversionAlgorithm] = {
        "direct": lambda hvp, x: np.linalg.solve(hvp(np.eye(n_params)), x.T).T,  # type: ignore
        "cg": lambda hvp, x: conjugate_gradient(LinearOperator((n_params, n_params), matvec=hvp), x, progress),  # type: ignore
        "batched_cg": lambda hvp, x: batched_preconditioned_conjugate_gradient(  # type: ignore
            hvp, x, **inversion_method_kwargs
        )[
            0
        ],
    }

    influence_factors = calculate_influence_factors(
        differentiable_model,
        x,
        y,
        x_test,
        y_test,
        dict_fact_algos[inversion_method],
        lam=hessian_regularization,
        progress=progress,
    )
    influence_function = influence_type_function_dict[influence_type]

    return influence_function(
        differentiable_model,
        x,
        y,
        influence_factors,
        progress,
    )



from enum import Enum


class Status(Enum):
    """Status of a computation.

    Statuses can be combined using bitwise or (``|``) and bitwise and (``&``) to
    get the status of a combined computation. For example, if we have two
    computations, one that has converged and one that has failed, then the
    combined status is ``Status.Converged | Status.Failed == Status.Converged``,
    but ``Status.Converged & Status.Failed == Status.Failed``.


    :OR:

    The result of bitwise or-ing two valuation statuses with ``|`` is given
    by the following table:

    +---+---+---+---+
    |   | P | C | F |
    +===+===+===+===+
    | P | P | C | P |
    +---+---+---+---+
    | C | C | C | C |
    +---+---+---+---+
    | F | P | C | F |
    +---+---+---+---+

    where P = Pending, C = Converged, F = Failed.

    :AND:

    The result of bitwise and-ing two valuation statuses with ``&`` is given
    by the following table:

    +---+---+---+---+
    |   | P | C | F |
    +===+===+===+===+
    | P | P | P | F |
    +---+---+---+---+
    | C | P | C | F |
    +---+---+---+---+
    | F | F | F | F |
    +---+---+---+---+

    where P = Pending, C = Converged, F = Failed.

    :NOT:

    The result of bitwise negation of a Status with ``~`` is ``Failed`` if
    the status is ``Converged``, or ``Converged`` otherwise:

        ~P == C, ~C == F, ~F == C

    :Boolean casting:

    A Status evaluates to ``True`` iff it's ``Converged`` or ``Failed``:

        bool(Status.Pending) == False
        bool(Status.Converged) == True
        bool(Status.Failed) == True

    .. warning::
       These truth values are **inconsistent** with the usual boolean operations.
       In particular the XOR of two instances of ``Status`` is not the same as
       the XOR of their boolean values.

    """

    Pending = "pending"
    Converged = "converged"
    Failed = "failed"

    def __or__(self, other: "Status") -> "Status":
        if self == Status.Converged or other == Status.Converged:
            return Status.Converged
        if self == Status.Pending or other == Status.Pending:
            return Status.Pending
        if self == Status.Failed and other == Status.Failed:
            return Status.Failed
        # Should be unreachable
        raise RuntimeError(f"Unexpected statuses: {self} and {other}")

    def __and__(self, other: "Status") -> "Status":
        # Careful, the order of tests matters here!
        if self == Status.Failed or other == Status.Failed:
            return Status.Failed
        if self == Status.Pending or other == Status.Pending:
            return Status.Pending
        if self == Status.Converged and other == Status.Converged:
            return Status.Converged
        # Should be unreachable
        raise RuntimeError(f"Unexpected statuses: {self} and {other}")

    def __invert__(self) -> "Status":
        if self == Status.Converged:
            return Status.Failed
        return Status.Converged

    def __bool__(self) -> bool:
        return self != Status.Pending



"""
This package holds all routines for the computation of Shapley Data value. Users
will want to use :func:`~pydvl.value.shapley.common.compute_shapley_values` as a
single interface to all methods defined in the modules.

Please refer to :ref:`data valuation` for an overview of Shapley Data value.
"""

from ..result import *
from ..stopping import *
from .common import *
from .gt import *
from .knn import *
from .montecarlo import *
from .naive import *
from .owen import *
from .truncated import *
from .types import *



from typing import Any, List, Optional, OrderedDict, Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from matplotlib.axes import Axes
from numpy.typing import NDArray
from scipy.stats import norm


def shaded_mean_std(
    data: np.ndarray,
    abscissa: Optional[Sequence[Any]] = None,
    num_std: float = 1.0,
    mean_color: Optional[str] = "dodgerblue",
    shade_color: Optional[str] = "lightblue",
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    ax: Optional[Axes] = None,
    **kwargs,
) -> Axes:
    """The usual mean +- x std deviations plot to aggregate runs of experiments.

    :param data: axis 0 is to be aggregated on (e.g. runs) and axis 1 is the
        data for each run.
    :param abscissa: values for the x axis. Leave empty to use increasing
        integers.
    :param num_std: number of standard deviations to shade around the mean.
    :param mean_color: color for the mean
    :param shade_color: color for the shaded region
    :param title:
    :param xlabel:
    :param ylabel:
    :param ax: If passed, axes object into which to insert the figure. Otherwise,
        a new figure is created and returned
    :param kwargs: these are forwarded to the ax.plot() call for the mean.

    :return: The axes used (or created)
    """
    assert len(data.shape) == 2
    mean = data.mean(axis=0)
    std = num_std * data.std(axis=0)

    if ax is None:
        fig, ax = plt.subplots()
    if abscissa is None:
        abscissa = list(range(data.shape[1]))

    ax.fill_between(abscissa, mean - std, mean + std, alpha=0.3, color=shade_color)
    ax.plot(abscissa, mean, color=mean_color, **kwargs)

    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ax


def shapley_results(results: dict, filename: str = None):
    """
    FIXME: change this to use dataframes

    :param results: dict
    :param filename: For plt.savefig(). Set to None to disable saving.

    Here's an example results dictionary::

        results = {
            "all_values": num_runs x num_points
            "backward_scores": num_runs x num_points,
            "backward_scores_reversed": num_runs x num_points,
            "backward_random_scores": num_runs x num_points,
            "forward_scores": num_runs x num_points,
            "forward_scores_reversed": num_runs x num_points,
            "forward_random_scores": num_runs x num_points,
            "max_iterations": int,
            "score_name" str,
            "num_points": int
        }
    """
    plt.figure(figsize=(16, 5))
    num_runs = len(results["all_values"])
    num_points = len(results["backward_scores"][0])
    use_points = int(0.6 * num_points)

    plt.subplot(1, 2, 1)
    values = np.array(results["backward_scores"])[:, :use_points]
    shaded_mean_std(values, color="b", label="By increasing shapley value")

    values = np.array(results["backward_scores_reversed"])[:, :use_points]
    shaded_mean_std(values, color="g", label="By decreasing shapley value")

    values = np.array(results["backward_random_scores"])[:, :use_points]
    shaded_mean_std(values, color="r", linestyle="--", label="At random")

    plt.ylabel(f'Score ({results.get("score_name")})')
    plt.xlabel("Points removed")
    plt.title(
        f"Effect of point removal. "
        f'MonteCarlo with {results.get("max_iterations")} iterations '
        f"over {num_runs} runs"
    )
    plt.legend()

    plt.subplot(1, 2, 2)

    values = np.array(results["forward_scores"])[:, :use_points]
    shaded_mean_std(values, color="b", label="By increasing shapley value")

    values = np.array(results["forward_scores_reversed"])[:, :use_points]
    shaded_mean_std(values, color="g", label="By decreasing shapley value")

    values = np.array(results["forward_random_scores"])[:, :use_points]
    shaded_mean_std(values, color="r", linestyle="--", label="At random")

    plt.ylabel(f'Score ({results.get("score_name")})')
    plt.xlabel("Points added")
    plt.title(
        f"Effect of point addition. "
        f'MonteCarlo with {results["max_iterations"]} iterations '
        f"over {num_runs} runs"
    )
    plt.legend()

    if filename:
        plt.savefig(filename, dpi=300)


def spearman_correlation(vv: List[OrderedDict], num_values: int, pvalue: float):
    """Simple matrix plots with spearman correlation for each pair in vv.

    :param vv: list of OrderedDicts with index: value. Spearman correlation
        is computed for the keys.
    :param num_values: Use only these many values from the data (from the start
        of the OrderedDicts)
    :param pvalue: correlation coefficients for which the p-value is below the
        threshold `pvalue/len(vv)` will be discarded.
    """
    r: np.ndarray = np.ndarray((len(vv), len(vv)))
    p: np.ndarray = np.ndarray((len(vv), len(vv)))
    for i, a in enumerate(vv):
        for j, b in enumerate(vv):
            from scipy.stats._stats_py import SpearmanrResult

            spearman: SpearmanrResult = sp.stats.spearmanr(
                list(a.keys())[:num_values], list(b.keys())[:num_values]
            )
            r[i][j] = (
                spearman.correlation if spearman.pvalue < pvalue / len(vv) else np.nan
            )  # Bonferroni correction
            p[i][j] = spearman.pvalue
    fig, axs = plt.subplots(1, 2, figsize=(16, 7))
    plot1 = axs[0].matshow(r, vmin=-1, vmax=1)
    axs[0].set_title(f"Spearman correlation (top {num_values} values)")
    axs[0].set_xlabel("Runs")
    axs[0].set_ylabel("Runs")
    fig.colorbar(plot1, ax=axs[0])
    plot2 = axs[1].matshow(p, vmin=0, vmax=1)
    axs[1].set_title("p-value")
    axs[1].set_xlabel("Runs")
    axs[1].set_ylabel("Runs")
    fig.colorbar(plot2, ax=axs[1])

    return fig


def plot_shapley(
    df: pd.DataFrame,
    *,
    level: float = 0.05,
    ax: plt.Axes = None,
    title: str = None,
    xlabel: str = None,
    ylabel: str = None,
) -> plt.Axes:
    """Plots the shapley values, as returned from
    :func:`~pydvl.value.shapley.common.compute_shapley_values`, with error bars
    corresponding to an $\alpha$-level confidence interval.

    :param df: dataframe with the shapley values
    :param level: confidence level for the error bars
    :param ax: axes to plot on or None if a new subplots should be created
    :param title: string, title of the plot
    :param xlabel: string, x label of the plot
    :param ylabel: string, y label of the plot
    :return: the axes created or used
    """
    if ax is None:
        _, ax = plt.subplots()

    yerr = norm.ppf(1 - level / 2) * df["data_value_stderr"]

    ax.errorbar(x=df.index, y=df["data_value"], yerr=yerr, fmt="o", capsize=6)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    plt.xticks(rotation=60)
    return ax


def plot_influence_distribution_by_label(
    influences: NDArray[np.float_], labels: NDArray[np.float_], title_extra: str = ""
):
    """Plots the histogram of the influence that all samples in the training set
     have over a single sample index, separated by labels.

    :param influences: array of influences (training samples x test samples)
    :param labels: labels for the training set.
    :param title_extra:
    """
    _, ax = plt.subplots()
    unique_labels = np.unique(labels)
    for label in unique_labels:
        ax.hist(influences[labels == label], label=label, alpha=0.7)
    ax.set_xlabel("Influence values")
    ax.set_ylabel("Number of samples")
    ax.set_title(f"Distribution of influences " + title_extra)
    ax.legend()
    plt.show()



"""
This module implements Group Testing for the approximation of Shapley values, as
introduced in :footcite:t:`jia_efficient_2019`. The sampling of index subsets is
done in such a way that an approximation to the true Shapley values can be
computed with guarantees.

.. warning::
   This method is very inefficient. Potential improvements to the
   implementation notwithstanding, convergence seems to be very slow (in terms
   of evaluations of the utility required). We recommend other Monte Carlo
   methods instead.

You can read more :ref:`in the documentation<data valuation>`.

.. versionadded:: 0.4.0

"""
import logging
from collections import namedtuple
from typing import Iterable, Tuple, TypeVar, cast

import cvxpy as cp
import numpy as np
from numpy.typing import NDArray

from pydvl.utils import MapReduceJob, ParallelConfig, Utility, maybe_progress
from pydvl.utils.numeric import random_subset_of_size
from pydvl.utils.parallel.backend import effective_n_jobs
from pydvl.utils.status import Status
from pydvl.value import ValuationResult

__all__ = ["group_testing_shapley", "num_samples_eps_delta"]

log = logging.getLogger(__name__)

T = TypeVar("T", NDArray[np.float_], float)
GTConstants = namedtuple("GTConstants", ["kk", "Z", "q", "q_tot", "T"])


def _constants(
    n: int, epsilon: float, delta: float, utility_range: float
) -> GTConstants:
    """A helper function returning the constants for the algorithm. Pretty ugly,
    yes.

    :param n: The number of data points.
    :param epsilon: The error tolerance.
    :param delta: The confidence level.
    :param utility_range: The range of the utility function.

    :return: A namedtuple with the constants. The fields are the same as in the
        paper:
        - kk: the sample sizes (i.e. an array of 1, 2, ..., n - 1)
        - Z: the normalization constant
        - q: the probability of drawing a sample of size k
        - q_tot: another normalization constant
        - T: the number of iterations. This will be -1 if the utility_range is
            infinite. E.g. because the :class:`~pydvl.utils.score.Scorer` does
            not define a range.
    """
    r = utility_range

    kk = np.arange(1, n)  # sample sizes
    Z = 2 * (1.0 / kk).sum()
    q = (1 / kk + 1 / (n - kk)) / Z
    q_tot = (n - 2) / n * q[0] + np.inner(
        q[1:], 1 + 2 * kk[1:] * (kk[1:] - n) / (n * (n - 1))
    )

    def h(u: T) -> T:
        return cast(T, (1 + u) * np.log(1 + u) - u)

    # The implementation in GitHub defines a different bound:
    # T_code = int( 4
    #     / (1 - q_tot**2)
    #     / h(2 * epsilon / Z / r / (1 - q_tot**2))
    #     * np.log(n * (n - 1) / (2 * delta))
    # )
    if r == np.inf:
        log.warning(
            "Group Testing: cannot estimate minimum number of iterations for "
            "unbounded utilities. Please specify a range in the Scorer if "
            "you need this estimate."
        )
        min_iter = -1
    else:
        min_iter = 8 * np.log(n * (n - 1) / (2 * delta)) / (1 - q_tot**2)
        min_iter /= h(2 * epsilon / (np.sqrt(n) * Z * r * (1 - q_tot**2)))

    return GTConstants(kk=kk, Z=Z, q=q, q_tot=q_tot, T=int(min_iter))


def num_samples_eps_delta(
    eps: float, delta: float, n: int, utility_range: float
) -> int:
    r"""Implements the formula in Theorem 3 of :footcite:t:`jia_efficient_2019`
    which gives a lower bound on the number of samples required to obtain an
    (ε/√n,δ/(N(N-1))-approximation to all pair-wise differences of Shapley
    values, wrt. $\ell_2$ norm.

    :param eps: ε
    :param delta: δ
    :param n: Number of data points
    :param utility_range: Range of the :class:`~pydvl.utils.utility.Utility`
        function
    :return: Number of samples from $2^{[n]}$ guaranteeing ε/√n-correct Shapley
        pair-wise differences of values with probability 1-δ/(N(N-1)).

    .. versionadded:: 0.4.0

    """
    constants = _constants(n=n, epsilon=eps, delta=delta, utility_range=utility_range)
    return int(constants.T)


def _group_testing_shapley(
    u: Utility, n_samples: int, progress: bool = False, job_id: int = 1
):
    """Helper function for :func:`group_testing_shapley`.

    Computes utilities of sets sampled using the strategy for estimating the
    differences in Shapley values.

    :param u: Utility object with model, data, and scoring function.
    :param n_samples: total number of samples (subsets) to use.
    :param progress: Whether to display progress bars for each job.
    :param job_id: id to use for reporting progress (e.g. to place progres bars)
    :return:
    """
    rng = np.random.default_rng()
    n = len(u.data.indices)
    const = _constants(n, 1, 1, 1)  # don't care about eps,delta,range

    betas = np.zeros(shape=(n_samples, n), dtype=np.int_)  # indicator vars
    uu = np.empty(n_samples)  # utilities

    for t in maybe_progress(n_samples, progress=progress, position=job_id):
        k = rng.choice(const.kk, size=1, p=const.q).item()
        s = random_subset_of_size(u.data.indices, k)
        uu[t] = u(s)
        betas[t, s] = 1
    return uu, betas


def group_testing_shapley(
    u: Utility,
    n_samples: int,
    epsilon: float,
    delta: float,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
    **options,
) -> ValuationResult:
    """Implements group testing for approximation of Shapley values as described
    in :footcite:t:`jia_efficient_2019`.

    .. warning::
       This method is very inefficient. It requires several orders of magnitude
       more evaluations of the utility than others in
       :mod:`~pydvl.value.shapley.montecarlo`. It also uses several intermediate
       objects like the results from the runners and the constraint matrices
       which can become rather large.

    By picking a specific distribution over subsets, the differences in Shapley
    values can be approximated with a Monte Carlo sum. These are then used to
    solve for the individual values in a feasibility problem.

    :param u: Utility object with model, data, and scoring function
    :param n_samples: Number of tests to perform. Use
        :func:`num_samples_eps_delta` to estimate this.
    :param epsilon: From the (ε,δ) sample bound. Use the same as for the
        estimation of ``n_iterations``.
    :param delta: From the (ε,δ) sample bound. Use the same as for the
        estimation of ``n_iterations``.
    :param n_jobs: Number of parallel jobs to use. Each worker performs a chunk
        of all tests (i.e. utility evaluations).
    :param config: Object configuring parallel computation, with cluster
        address, number of cpus, etc.
    :param progress: Whether to display progress bars for each job.
    :param options: Additional options to pass to `cvxpy.Problem.solve()
        <https://www.cvxpy.org/tutorial/advanced/index.html#solve-method-options>`_.
        E.g. to change the solver (which defaults to `cvxpy.SCS`) pass
        `solver=cvxpy.CVXOPT`.

    :return: Object with the data values.

    .. versionadded:: 0.4.0

    .. versionchanged:: 0.5.0
       Changed the solver to cvxpy instead of scipy's linprog. Added the ability
       to pass arbitrary options to it.
    """

    n = len(u.data.indices)

    const = _constants(
        n=n,
        epsilon=epsilon,
        delta=delta,
        utility_range=u.score_range.max() - u.score_range.min(),
    )
    T = n_samples
    if T < const.T:
        log.warning(
            f"n_samples of {T} are below the required {const.T} for the "
            f"ε={epsilon:.02f} guarantee at δ={1 - delta:.02f} probability"
        )

    samples_per_job = max(1, n_samples // effective_n_jobs(n_jobs, config))

    def reducer(
        results_it: Iterable[Tuple[NDArray, NDArray]]
    ) -> Tuple[NDArray, NDArray]:
        return np.concatenate(list(x[0] for x in results_it)).astype(
            np.float_
        ), np.concatenate(list(x[1] for x in results_it)).astype(np.int_)

    map_reduce_job: MapReduceJob[Utility, Tuple[NDArray, NDArray]] = MapReduceJob(
        u,
        map_func=_group_testing_shapley,
        reduce_func=reducer,
        map_kwargs=dict(n_samples=samples_per_job, progress=progress),
        config=config,
        n_jobs=n_jobs,
    )
    uu, betas = map_reduce_job()

    # Matrix of estimated differences. See Eqs. (3) and (4) in the paper.
    C = np.zeros(shape=(n, n))
    for i in range(n):
        for j in range(i + 1, n):
            C[i, j] = np.dot(uu, betas[:, i] - betas[:, j])
    C *= const.Z / T
    total_utility = u(u.data.indices)

    ###########################################################################
    # Solution of the constraint problem with cvxpy

    v = cp.Variable(n)
    constraints = [cp.sum(v) == total_utility]
    for i in range(n):
        for j in range(i + 1, n):
            constraints.append(v[i] - v[j] <= epsilon + C[i, j])
            constraints.append(v[j] - v[i] <= epsilon - C[i, j])

    problem = cp.Problem(cp.Minimize(0), constraints)
    solver = options.pop("solver", cp.SCS)
    problem.solve(solver=solver, **options)

    if problem.status != "optimal":
        log.warning(f"cvxpy returned status {problem.status}")
        values = (
            np.nan * np.ones_like(u.data.indices)
            if not hasattr(v.value, "__len__")
            else v.value
        )
        status = Status.Failed
    else:
        values = v.value
        status = Status.Converged

    return ValuationResult(
        algorithm="group_testing_shapley",
        status=status,
        values=values,
        data_names=u.data.data_names,
        solver_status=problem.status,
    )



import numpy as np

from pydvl.utils import Utility, maybe_progress
from pydvl.utils.status import Status
from pydvl.value.result import ValuationResult

__all__ = ["naive_loo"]


def naive_loo(u: Utility, *, progress: bool = True) -> ValuationResult:
    r"""Computes leave one out value:

    $$v(i) = u(D) - u(D \setminus \{i\}) $$

    :param u: Utility object with model, data, and scoring function
    :param progress: If True, display a progress bar
    :return: Object with the data values.
    """

    if len(u.data) < 3:
        raise ValueError("Dataset must have at least 2 elements")

    values = np.zeros_like(u.data.indices, dtype=np.float_)
    all_indices = set(u.data.indices)
    total_utility = u(u.data.indices)
    for i in maybe_progress(u.data.indices, progress):  # type: ignore
        subset = all_indices.difference({i})
        values[i] = total_utility - u(subset)

    return ValuationResult(
        algorithm="naive_loo",
        status=Status.Converged,
        values=values,
        data_names=u.data.data_names,
    )



"""
This module contains routines for numerical computations used across the
library.
"""
from __future__ import annotations

from itertools import chain, combinations
from typing import Collection, Generator, Iterator, Optional, Tuple, TypeVar, overload

import numpy as np
from numpy.typing import NDArray

__all__ = [
    "running_moments",
    "linear_regression_analytical_derivative_d2_theta",
    "linear_regression_analytical_derivative_d_theta",
    "linear_regression_analytical_derivative_d_x_d_theta",
    "num_samples_permutation_hoeffding",
    "powerset",
    "random_matrix_with_condition_number",
    "random_subset",
    "random_powerset",
    "random_subset_of_size",
    "top_k_value_accuracy",
]

T = TypeVar("T", bound=np.generic)


def powerset(s: NDArray[T]) -> Iterator[Collection[T]]:
    """Returns an iterator for the power set of the argument.

     Subsets are generated in sequence by growing size. See
     :func:`random_powerset` for random sampling.

    >>> import numpy as np
    >>> from pydvl.utils.numeric import powerset
    >>> list(powerset(np.array((1,2))))
    [(), (1,), (2,), (1, 2)]

     :param s: The set to use
     :return: An iterator
     :raises TypeError: If the argument is not an ``Iterable``.
    """
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))


def num_samples_permutation_hoeffding(eps: float, delta: float, u_range: float) -> int:
    """Lower bound on the number of samples required for MonteCarlo Shapley to
    obtain an (ε,δ)-approximation.

    That is: with probability 1-δ, the estimated value for one data point will
    be ε-close to the true quantity, if at least this many permutations are
    sampled.

    :param eps: ε > 0
    :param delta: 0 < δ <= 1
    :param u_range: Range of the :class:`~pydvl.utils.utility.Utility` function
    :return: Number of _permutations_ required to guarantee ε-correct Shapley
        values with probability 1-δ
    """
    return int(np.ceil(np.log(2 / delta) * 2 * u_range**2 / eps**2))


def random_subset(s: NDArray[T], q: float = 0.5) -> NDArray[T]:
    """Returns one subset at random from ``s``.

    :param s: set to sample from
    :param q: Sampling probability for elements. The default 0.5 yields a
        uniform distribution over the power set of s.
    :return: the subset
    """
    rng = np.random.default_rng()
    selection = rng.uniform(size=len(s)) > q
    return s[selection]


def random_powerset(
    s: NDArray[T], n_samples: Optional[int] = None, q: float = 0.5
) -> Generator[NDArray[T], None, None]:
    """Samples subsets from the power set of the argument, without
    pre-generating all subsets and in no order.

    See `powerset()` if you wish to deterministically generate all subsets.

    To generate subsets, `len(s)` Bernoulli draws with probability `q` are
    drawn. The default value of `q = 0.5` provides a uniform distribution over
    the power set of `s`. Other choices can be used e.g. to implement
    :func:`Owen sampling
    <pydvl.value.shapley.montecarlo.owen_sampling_shapley>`.

    :param s: set to sample from
    :param n_samples: if set, stop the generator after this many steps.
        Defaults to `np.iinfo(np.int32).max`
    :param q: Sampling probability for elements. The default 0.5 yields a
        uniform distribution over the power set of s.

    :return: Samples from the power set of s
    :raises: TypeError: if the data `s` is not a NumPy array
    :raises: ValueError: if the element sampling probability is not in [0,1]

    """
    if not isinstance(s, np.ndarray):
        raise TypeError("Set must be an NDArray")
    if q < 0 or q > 1:
        raise ValueError("Element sampling probability must be in [0,1]")

    total = 1
    if n_samples is None:
        n_samples = np.iinfo(np.int32).max
    while total <= n_samples:
        yield random_subset(s, q)
        total += 1


def random_subset_of_size(s: NDArray[T], size: int) -> NDArray[T]:
    """Samples a random subset of given size uniformly from the powerset
    of ``s``.

    :param s: Set to sample from
    :param size: Size of the subset to generate
    :return: The subset
    :raises ValueError: If size > len(s)
    """
    if size > len(s):
        raise ValueError("Cannot sample subset larger than set")
    rng = np.random.default_rng()
    return rng.choice(s, size=size, replace=False)


def random_matrix_with_condition_number(n: int, condition_number: float) -> "NDArray":
    """Constructs a square matrix with a given condition number.

    Taken from:
    https://gist.github.com/bstellato/23322fe5d87bb71da922fbc41d658079#file-random_mat_condition_number-py

    Also see:
    https://math.stackexchange.com/questions/1351616/condition-number-of-ata.

    :param n: size of the matrix
    :param condition_number: duh
    :return: An (n,n) matrix with the requested condition number.
    """
    if n < 2:
        raise ValueError("Matrix size must be at least 2")

    if condition_number <= 1:
        raise ValueError("Condition number must be greater than 1")

    log_condition_number = np.log(condition_number)
    exp_vec = np.arange(
        -log_condition_number / 4.0,
        log_condition_number * (n + 1) / (4 * (n - 1)),
        log_condition_number / (2.0 * (n - 1)),
    )
    exp_vec = exp_vec[:n]
    s: np.ndarray = np.exp(exp_vec)
    S = np.diag(s)
    U, _ = np.linalg.qr((np.random.rand(n, n) - 5.0) * 200)
    V, _ = np.linalg.qr((np.random.rand(n, n) - 5.0) * 200)
    P: np.ndarray = U.dot(S).dot(V.T)
    P = P.dot(P.T)
    return P


def linear_regression_analytical_derivative_d_theta(
    linear_model: Tuple["NDArray", "NDArray"], x: "NDArray", y: "NDArray"
) -> "NDArray":
    """
    :param linear_model: A tuple of np.ndarray' of shape [NxM] and [N] representing A and b respectively.
    :param x: A np.ndarray of shape [BxM].
    :param y: A np.nparray of shape [BxN].
    :returns: A np.ndarray of shape [Bx((N+1)*M)], where each row vector is [d_theta L(x, y), d_b L(x, y)]
    """

    A, b = linear_model
    n, m = list(A.shape)
    residuals = x @ A.T + b - y
    kron_product = np.expand_dims(residuals, axis=2) * np.expand_dims(x, axis=1)
    test_grads = np.reshape(kron_product, [-1, n * m])
    full_grads = np.concatenate((test_grads, residuals), axis=1)
    return full_grads / n  # type: ignore


def linear_regression_analytical_derivative_d2_theta(
    linear_model: Tuple["NDArray", "NDArray"], x: "NDArray", y: "NDArray"
) -> "NDArray":
    """
    :param linear_model: A tuple of np.ndarray' of shape [NxM] and [N] representing A and b respectively.
    :param x: A np.ndarray of shape [BxM],
    :param y: A np.nparray of shape [BxN].
    :returns: A np.ndarray of shape [((N+1)*M)x((N+1)*M)], representing the Hessian. It gets averaged over all samples.
    """
    A, b = linear_model
    n, m = tuple(A.shape)
    d2_theta = np.einsum("ia,ib->iab", x, x)
    d2_theta = np.mean(d2_theta, axis=0)
    d2_theta = np.kron(np.eye(n), d2_theta)
    d2_b = np.eye(n)
    mean_x = np.mean(x, axis=0, keepdims=True)
    d_theta_d_b = np.kron(np.eye(n), mean_x)
    top_matrix = np.concatenate((d2_theta, d_theta_d_b.T), axis=1)
    bottom_matrix = np.concatenate((d_theta_d_b, d2_b), axis=1)
    full_matrix = np.concatenate((top_matrix, bottom_matrix), axis=0)
    return full_matrix / n  # type: ignore


def linear_regression_analytical_derivative_d_x_d_theta(
    linear_model: Tuple["NDArray", "NDArray"], x: "NDArray", y: "NDArray"
) -> "NDArray":
    """
    :param linear_model: A tuple of np.ndarray of shape [NxM] and [N] representing A and b respectively.
    :param x: A np.ndarray of shape [BxM].
    :param y: A np.nparray of shape [BxN].
    :returns: A np.ndarray of shape [Bx((N+1)*M)xM], representing the derivative.
    """

    A, b = linear_model
    N, M = tuple(A.shape)
    residuals = x @ A.T + b - y
    B = len(x)
    outer_product_matrix = np.einsum("ab,ic->iacb", A, x)
    outer_product_matrix = np.reshape(outer_product_matrix, [B, M * N, M])
    tiled_identity = np.tile(np.expand_dims(np.eye(M), axis=0), [B, N, 1])
    outer_product_matrix += tiled_identity * np.expand_dims(
        np.repeat(residuals, M, axis=1), axis=2
    )
    b_part_derivative = np.tile(np.expand_dims(A, axis=0), [B, 1, 1])
    full_derivative = np.concatenate((outer_product_matrix, b_part_derivative), axis=1)
    return full_derivative / N  # type: ignore


@overload
def running_moments(
    previous_avg: float, previous_variance: float, count: int, new_value: float
) -> Tuple[float, float]:
    ...


@overload
def running_moments(
    previous_avg: NDArray[np.float_],
    previous_variance: NDArray[np.float_],
    count: int,
    new_value: NDArray[np.float_],
) -> Tuple[NDArray[np.float_], NDArray[np.float_]]:
    ...


def running_moments(
    previous_avg: float | NDArray[np.float_],
    previous_variance: float | NDArray[np.float_],
    count: int,
    new_value: float | NDArray[np.float_],
) -> Tuple[float | NDArray[np.float_], float | NDArray[np.float_]]:
    """Uses Welford's algorithm to calculate the running average and variance of
     a set of numbers.

    See `Welford's algorithm in wikipedia
    <https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm>`_

    .. warning::
       This is not really using Welford's correction for numerical stability
       for the variance. (FIXME)

    .. todo::
       This could be generalised to arbitrary moments. See `this paper
       <https://www.osti.gov/biblio/1028931>`_


    :param previous_avg: average value at previous step
    :param previous_variance: variance at previous step
    :param count: number of points seen so far
    :param new_value: new value in the series of numbers
    :return: new_average, new_variance, calculated with the new count
    """
    # broadcasted operations seem not to be supported by mypy, so we ignore the type
    new_average = (new_value + count * previous_avg) / (count + 1)  # type: ignore
    new_variance = previous_variance + (
        (new_value - previous_avg) * (new_value - new_average) - previous_variance
    ) / (count + 1)
    return new_average, new_variance


def top_k_value_accuracy(
    y_true: NDArray[np.float_], y_pred: NDArray[np.float_], k: int = 3
) -> float:
    """Computes the top-k accuracy for the estimated values by comparing indices
    of the highest k values.

    :param y_true: Exact/true value
    :param y_pred: Predicted/estimated value
    :param k: Number of the highest values taken into account
    :return: Accuracy
    """
    top_k_exact_values = np.argsort(y_true)[-k:]
    top_k_pred_values = np.argsort(y_pred)[-k:]
    top_k_accuracy = len(np.intersect1d(top_k_exact_values, top_k_pred_values)) / k
    return top_k_accuracy
